/*
    control++ : System configuration tool
    Copyright (C) 2017-2021 Alexey Appolonov

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "main.h"
#include "controller.h"

//-----------------------------------------------------------------------------
// Display help notes

void PrintHelp(const string &base){
    cout << "Set control++ macro mode:" << endl
         << "\t" << base << " <mode_name>" << endl
         << "Set control++ micro mode for the unit:" << endl
         << "\t" << base << " <unit_name> <mode_name>" << endl
         << "Restore state that held before macro mode setting:" << endl
         << "\t" << base << " --reset" << endl
         << "Restore related state that held before micro mode setting:" << endl
         << "\t" << base << " <unit_name> --reset" << endl
         << "Display list of available modes:" << endl
         << "\t" << base << " --list" << endl
         << "Display current macro and micro modes and check compliance "
            "of the macro mode:" << endl
         << "\t" << base << " --status" << endl
         << "Display current macro and micro modes and check compliance "
            "of micro mode of the unit:" << endl
         << "\t" << base << " <unit_name> --status" << endl
         << "Display path and contents of config file:" << endl
         << "\t" << base << " --conf" << endl
         << "Show this help notes:" << endl
         << "\t" << base << " --help" << endl
         << "Use '--plain | -p' flag to enable plain output mode" << endl
         << "Use '--force | -f' flag to force setting of the mode "
         << "(disables approval dialogs)" << endl
         << "Use '--nonuniq | -n' flag to disable detection of non-unique "
         << "permission modes" << endl
         << "Use '--verbose | -v' flag for verbose output" << endl;
    }

//-----------------------------------------------------------------------------

TRunMode run_mode;
TPrinter printer;

int main(int argc, char *argv[]){

    // Main control unit, all calls to the subunits addressed through it
    TController controller;

    // Converting command line arguments from C-arrays to strings
    const string exe = argv[0];
    const auto params = [argc, argv](void){
        vector<string> v;
        for(int i = 1; i < argc; i++){
            v.push_back(argv[i]);
            }
        return v;
        }();

    // Making abs path of the program available to every unit
    char *pwd = get_current_dir_name();
    if(!pwd){
        FatalError("Can't get name of the current directory");
        }
    run_mode.WriteExePath(pwd + StripPath(exe));

    string unit, mode;
    bool wrong_arg = false;

    #define F_HELP   "--help"
    #define F_CONF   "--conf"
    #define F_DOM    "--dom"
    #define F_LIST   "--list"
    #define F_STATUS "--status"
    #define F_RESET  "--reset"

    map<string, bool> flags = {
        {F_HELP, false},
        {F_CONF, false},
        {F_DOM, false},
        {F_LIST, false},
        {F_STATUS, false},
        {F_RESET, false}
        };

    // Set given flag
    auto set_flag = [&](const string &flag){
        // Only one flag can be set at a time
        for(const auto &kv : flags){
            if(kv.first != flag && kv.second == true){
                return false;
                }
            }
        if(flag == F_STATUS || flag == F_RESET){
            if(!unit.empty()){
                return false;
                }
            unit.swap(mode);
            }
        flags[flag] = true;
        return true;
        };

    // Get state of a specified flag
    auto get_flag = [&](const string &flag){
        const auto it = flags.find(flag);
        if(it == flags.end()){
            return false;
            }
        return it->second;
        };

    // Parsing the flags
    for(const auto &param : params){
        if(flags.find(param) != flags.end()){
            if(!set_flag(param)){
                wrong_arg = true;
                break;
                }
            }
        else if(param == "--plain" || param == "-p"){
            printer.Plain(true);
            }
        else if(param == "--force" || param == "-f"){
            run_mode.WriteForce(true);
            }
        else if(param == "--nonuniq" || param == "-n"){
            run_mode.WriteFlag(M_NON_UNIQUE_PERM, false);
            }
        else if(param == "--verbose" || param == "-v"){
            run_mode.WriteFlag(M_VERBOSE, true);
            }
        else if(mode.empty() || unit.empty()){
            if(get_flag(F_STATUS) || get_flag(F_RESET)){
                if(!mode.empty() || !unit.empty()){
                    wrong_arg = true;
                    break;
                    }
                unit = param;
                }
            else if(mode.empty()){
                mode = param;
                }
            else{
                unit = mode;
                mode = param;
                }
            }
        else{
            wrong_arg = true;
            break;
            }
        }

    // Resetting a current mode equals setting a "reset" mode
    if(get_flag(F_RESET)){
        mode = RESET_MODE;
        }

    // Flags cannot be set twice
    run_mode.WriteFlag(M_NON_UNIQUE_PERM, true);
    run_mode.WriteFlag(M_VERBOSE, false);

    // Display help notes and exit
    if(params.empty() || get_flag(F_HELP) || wrong_arg){
        PrintHelp(exe);
        return wrong_arg ? 1 : 0;
        }

    // Display config and exit
    if(get_flag(F_CONF)){
        controller.PrintConfFile();
        return 0;
        }

    // Configure the main unit - if we passed through here
    // than we need to configure the program
    if(!controller.Configure()){
        printer.LineEnd("control++ can not be run without configuration file");
        return 1;
        }

    // Display configuration file DOM and exit
    if(get_flag(F_DOM)){
        controller.PrintConfDOM();
        }

    // Display list of all modes and exit
    else if(get_flag(F_LIST)){
        controller.PrintAvailableMacroAndMicroModes();
        }

    // Display current mode and exit
    else if(get_flag(F_STATUS)){
        controller.PrintCurrentModesOfTheUnits();
        if(controller.ConfigsControversy()){
            cout << "Mode description was changed since the last mode setting"
                 << endl;
            if(run_mode.Force()){
                printer.LineEnd("Acknowledging the '--force' flag, "
                    "checking mode compliance according to the modified "
                    "mode description");
                }
            else if(!Approval("Continue to check mode compliance according "
                        "to the modified mode description [y/N]? ", "y")){
                return 1;
                }
            }
        if(!controller.CheckModeCompliance("", unit)){
            return 1;
            }
        }

    // Set given mode
    else{
        if(!controller.SetMode(mode, unit)){
            printer.LineEnd("Set mode failed");
            return 1;
            }
        }

    return 0;
    }

//-----------------------------------------------------------------------------
